/*
 * Twidere - Twitter client for Android
 *
 *  Copyright (C) 2012-2015 Mariotaku Lee <mariotaku.lee@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.mariotaku.twidere.adapter

import android.annotation.SuppressLint
import android.arch.paging.AsyncPagedListDiffer
import android.arch.paging.PagedList
import android.arch.paging.setPagedListListener
import android.content.Context
import android.support.v7.recyclerview.extensions.AsyncDifferConfig
import android.support.v7.widget.RecyclerView
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.Space
import android.widget.TextView
import com.bumptech.glide.RequestManager
import org.mariotaku.kpreferences.get
import org.mariotaku.ktextension.contains
import org.mariotaku.ktextension.rangeOfSize
import org.mariotaku.ktextension.toInt
import org.mariotaku.ktextension.weak
import org.mariotaku.microblog.library.model.twitter.Activity.Action
import org.mariotaku.twidere.R
import org.mariotaku.twidere.adapter.callback.ItemCountsAdapterListUpdateCallback
import org.mariotaku.twidere.adapter.iface.IActivitiesAdapter
import org.mariotaku.twidere.adapter.iface.IGapSupportedAdapter
import org.mariotaku.twidere.adapter.iface.IItemCountsAdapter
import org.mariotaku.twidere.annotation.LoadMorePosition
import org.mariotaku.twidere.annotation.TimelineStyle
import org.mariotaku.twidere.constant.RecyclerViewTypes
import org.mariotaku.twidere.constant.newDocumentApiKey
import org.mariotaku.twidere.exception.UnsupportedCountIndexException
import org.mariotaku.twidere.extension.model.activityStatus
import org.mariotaku.twidere.model.ItemCounts
import org.mariotaku.twidere.model.ObjectId
import org.mariotaku.twidere.model.ParcelableActivity
import org.mariotaku.twidere.model.ParcelableMedia
import org.mariotaku.twidere.model.placeholder.ParcelableActivityPlaceholder
import org.mariotaku.twidere.model.placeholder.PlaceholderObject
import org.mariotaku.twidere.singleton.PreferencesSingleton
import org.mariotaku.twidere.util.IntentUtils
import org.mariotaku.twidere.util.OnLinkClickHandler
import org.mariotaku.twidere.util.TwidereLinkify
import org.mariotaku.twidere.util.paging.DiffItemCallbacks
import org.mariotaku.twidere.view.holder.ActivityTitleSummaryViewHolder
import org.mariotaku.twidere.view.holder.EmptyViewHolder
import org.mariotaku.twidere.view.holder.GapViewHolder
import org.mariotaku.twidere.view.holder.LoadIndicatorViewHolder
import org.mariotaku.twidere.view.holder.iface.IStatusViewHolder
import org.mariotaku.twidere.view.holder.status.StatusViewHolder
import java.util.*

class ParcelableActivitiesAdapter(
        context: Context,
        requestManager: RequestManager
) : LoadMoreSupportAdapter<RecyclerView.ViewHolder>(context, requestManager), IActivitiesAdapter,
        IItemCountsAdapter {

    override val itemCounts: ItemCounts = ItemCounts(2)

    override var loadMoreIndicatorPosition: Int
        get() = super.loadMoreIndicatorPosition
        set(value) {
            super.loadMoreIndicatorPosition = value
            updateItemCounts()
        }

    override var loadMoreSupportedPosition: Int
        get() = super.loadMoreSupportedPosition
        set(value) {
            super.loadMoreSupportedPosition = value
            updateItemCounts()
        }

    override val activityEventListener: IActivitiesAdapter.ActivityEventListener?
        get() = eventListener

    override val gapClickListener: IGapSupportedAdapter.GapClickListener?
        get() = eventListener

    override val mediaPreviewStyle: Int
        get() = statusAdapterDelegate.mediaPreviewStyle

    override val useStarsForLikes: Boolean
        get() = statusAdapterDelegate.useStarsForLikes

    override val mediaPreviewEnabled: Boolean
        get() = statusAdapterDelegate.mediaPreviewEnabled

    override val lightFont: Boolean
        get() = statusAdapterDelegate.lightFont

    override var showAccountsColor: Boolean
        get() = statusAdapterDelegate.showAccountsColor
        set(value) {
            statusAdapterDelegate.showAccountsColor = value
            notifyDataSetChanged()
        }

    val isNameFirst: Boolean
        get() = statusAdapterDelegate.nameFirst

    val activityStartIndex: Int
        get() = getItemStartPosition(ITEM_INDEX_ACTIVITY)

    var activities: PagedList<ParcelableActivity>?
        get() = pagedActivitiesHelper.currentList
        set(value) {
            pagedActivitiesHelper.submitList(value)
            gapLoadingIds.clear()
        }

    var activityClickListener: ActivityAdapterListener? = null

    var pagedListListener: ((list: PagedList<ParcelableActivity>?) -> Unit)? = null
        set(value) {
            field = value
            pagedActivitiesHelper.setPagedListListener(value)
        }

    private val inflater = LayoutInflater.from(context)
    private val twidereLinkify = TwidereLinkify(OnLinkClickHandler(context, null, PreferencesSingleton.get(this.context)))
    private val statusAdapterDelegate = DummyItemAdapter(context, twidereLinkify, this, requestManager)
    private val eventListener: EventListener
    private val gapLoadingIds: MutableSet<ObjectId<String>> = HashSet()

    private var pagedActivitiesHelper = AsyncPagedListDiffer<ParcelableActivity>(ItemCountsAdapterListUpdateCallback(this, ITEM_INDEX_ACTIVITY),
            AsyncDifferConfig.Builder<ParcelableActivity>(DiffItemCallbacks.activity).build())

    init {
        eventListener = EventListener(this)
        statusAdapterDelegate.updateOptions()
        setHasStableIds(true)
    }

    override fun getItemId(position: Int): Long {
        val countIndex = itemCounts.getItemCountIndex(position)
        return when (countIndex) {
            ITEM_INDEX_ACTIVITY -> {
                getRowId(position)
            }
            else -> {
                (countIndex.toLong() shl 32) or getItemViewType(position).toLong()
            }
        }
    }

    override fun getActivity(position: Int): ParcelableActivity {
        return getActivityInternal(position = position)
    }

    override fun getActivityCount(): Int {
        return itemCounts[ITEM_INDEX_ACTIVITY]
    }

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): RecyclerView.ViewHolder {
        when (viewType) {
            in RecyclerViewTypes.STATUS_TYPES -> {
                val holder = ParcelableStatusesAdapter.createStatusViewHolder(statusAdapterDelegate,
                        inflater, parent, TimelineStyle.PLAIN, viewType)
                return holder as RecyclerView.ViewHolder
            }
            RecyclerViewTypes.ACTIVITY_TITLE_SUMMARY -> {
                val view = inflater.inflate(R.layout.list_item_activity_summary_compact, parent, false)
                val holder = ActivityTitleSummaryViewHolder(view, this)
                holder.setupViewOptions()
                return holder
            }
            RecyclerViewTypes.GAP -> {
                val view = inflater.inflate(GapViewHolder.layoutResource, parent, false)
                return GapViewHolder(view)
            }
            RecyclerViewTypes.LOAD_INDICATOR -> {
                val view = inflater.inflate(R.layout.list_item_load_indicator, parent, false)
                return LoadIndicatorViewHolder(view)
            }
            RecyclerViewTypes.ACTIVITY_STUB -> {
                val view = inflater.inflate(R.layout.list_item_two_line, parent, false)
                return StubViewHolder(view)
            }
            RecyclerViewTypes.EMPTY -> {
                return EmptyViewHolder(Space(context))
            }
        }
        throw UnsupportedOperationException("Unsupported viewType $viewType")
    }


    override fun onBindViewHolder(holder: RecyclerView.ViewHolder, position: Int) {
        when (holder.itemViewType) {
            in RecyclerViewTypes.STATUS_TYPES -> {
                val activity = getActivityInternal(true, position)
                (holder as IStatusViewHolder).display(activity, displayInReplyTo = true)
            }
            RecyclerViewTypes.ACTIVITY_TITLE_SUMMARY -> {
                val activity = getActivityInternal(true, position)
                (holder as ActivityTitleSummaryViewHolder).display(activity)
            }
            RecyclerViewTypes.ACTIVITY_STUB -> {
                val activity = getActivityInternal(true, position)
                (holder as StubViewHolder).display(activity)
            }
            RecyclerViewTypes.GAP -> {
                val activity = getActivityInternal(true, position)
                val loading = gapLoadingIds.any { it.accountKey == activity.account_key && it.id == activity.id }
                (holder as GapViewHolder).display(loading)
            }
            RecyclerViewTypes.EMPTY -> {
            }
        }
    }

    override fun onViewAttachedToWindow(holder: RecyclerView.ViewHolder) {
        when (holder.itemViewType) {
            in RecyclerViewTypes.STATUS_TYPES -> {
                holder as StatusViewHolder
                holder.adapter = statusAdapterDelegate
                holder.setStatusClickListener(eventListener)
            }
            RecyclerViewTypes.ACTIVITY_TITLE_SUMMARY -> {
                holder as ActivityTitleSummaryViewHolder
                holder.setActivityEventListener(eventListener)
            }
            RecyclerViewTypes.GAP -> {
                holder as GapViewHolder
                holder.clickListener = eventListener
            }
        }
    }

    override fun getItemViewType(position: Int): Int {
        val countIndex = getItemCountIndex(position)
        when (countIndex) {
            ITEM_INDEX_ACTIVITY -> {
                val activity = getActivity(position)
                return when {
                    activity is PlaceholderObject -> RecyclerViewTypes.ACTIVITY_TITLE_SUMMARY
                    activity.is_gap -> RecyclerViewTypes.GAP
                    activity.is_filtered -> RecyclerViewTypes.EMPTY
                    else -> when (activity.action) {
                        Action.MENTION, Action.QUOTE, Action.REPLY -> ParcelableStatusesAdapter.statusItemViewType(activity)
                        Action.FOLLOW, Action.FAVORITE, Action.RETWEET,
                        Action.FAVORITED_RETWEET, Action.RETWEETED_RETWEET,
                        Action.RETWEETED_MENTION, Action.FAVORITED_MENTION,
                        Action.LIST_CREATED, Action.LIST_MEMBER_ADDED,
                        Action.MEDIA_TAGGED, Action.RETWEETED_MEDIA_TAGGED,
                        Action.FAVORITED_MEDIA_TAGGED, Action.JOINED_TWITTER -> RecyclerViewTypes.ACTIVITY_TITLE_SUMMARY
                        else -> RecyclerViewTypes.ACTIVITY_STUB
                    }
                }
            }
            ITEM_INDEX_LOAD_MORE_INDICATOR -> {
                return RecyclerViewTypes.LOAD_INDICATOR
            }
            else -> throw UnsupportedCountIndexException(countIndex, position)
        }
    }

    override fun addGapLoadingId(id: ObjectId<String>) {
        gapLoadingIds.add(id)
    }

    override fun removeGapLoadingId(id: ObjectId<String>) {
        gapLoadingIds.remove(id)
    }

    override fun getItemCount(): Int {
        return itemCounts.itemCount
    }

    override fun updateItemCounts() {
        itemCounts[1] = (LoadMorePosition.END in loadMoreIndicatorPosition).toInt()
    }

    fun findPositionBySortTimestamp(timestamp: Long): Int {
        if (timestamp <= 0) return RecyclerView.NO_POSITION
        val range = rangeOfSize(activityStartIndex, getActivityCount())
        if (range.isEmpty()) return RecyclerView.NO_POSITION
        if (timestamp < getTimestamp(range.last)) {
            return range.last
        }
        return range.indexOfFirst { timestamp >= getTimestamp(it) }
    }

    fun getTimestamp(adapterPosition: Int): Long {
        return getActivity(adapterPosition).timestamp
    }

    fun getRowId(adapterPosition: Int): Long {
        return getActivity(adapterPosition)._id
    }

    private fun getActivityInternal(loadAround: Boolean = false, position: Int): ParcelableActivity {
        val dataPosition = position - activityStartIndex
        val activityCount = getActivityCount()
        if (dataPosition < 0 || dataPosition >= activityCount) {
            val validRange = rangeOfSize(activityStartIndex, getActivityCount())
            throw IndexOutOfBoundsException("index: $position, valid range is $validRange")
        }
        return if (loadAround) {
            pagedActivitiesHelper.getItem(dataPosition)
        } else {
            pagedActivitiesHelper.currentList?.get(dataPosition)
        } ?: ParcelableActivityPlaceholder
    }

    interface ActivityAdapterListener {
        fun onGapClick(holder: GapViewHolder, position: Int)

        fun onActivityClick(holder: ActivityTitleSummaryViewHolder, position: Int)

        fun onStatusActionClick(holder: IStatusViewHolder, id: Int, position: Int)

        fun onStatusActionLongClick(holder: IStatusViewHolder, id: Int, position: Int): Boolean

        fun onStatusMenuClick(holder: IStatusViewHolder, menuView: View, position: Int)

        fun onMediaClick(holder: IStatusViewHolder, view: View, media: ParcelableMedia, position: Int)

        fun onStatusClick(holder: IStatusViewHolder, position: Int)

        fun onStatusLongClick(holder: IStatusViewHolder, position: Int): Boolean

        fun onQuotedStatusClick(holder: IStatusViewHolder, position: Int)

    }

    internal class StubViewHolder(itemView: View) : RecyclerView.ViewHolder(itemView) {

        internal val text1 = itemView.findViewById<TextView>(android.R.id.text1)
        internal val text2 = itemView.findViewById<TextView>(android.R.id.text2)

        init {
            text2.setSingleLine(false)
        }

        @SuppressLint("SetTextI18n")
        fun display(activity: ParcelableActivity) {
            text1.text = text1.resources.getString(R.string.unsupported_activity_action_title,
                    activity.action)
            text2.text = "host: ${activity.account_key.host}, id: ${activity.id}\n"
            text2.append(itemView.context.getString(R.string.unsupported_activity_action_summary))
        }
    }

    internal class EventListener(adapter: ParcelableActivitiesAdapter) :
            IStatusViewHolder.StatusClickListener, IGapSupportedAdapter.GapClickListener,
            IActivitiesAdapter.ActivityEventListener {

        val adapterRef by weak(adapter)

        override fun onGapClick(holder: GapViewHolder, position: Int) {
            val adapter = adapterRef ?: return
            val activity = adapter.getActivity(position)
            adapter.addGapLoadingId(ObjectId(activity.account_key, activity.id))
            adapter.activityClickListener?.onGapClick(holder, position)
        }

        override fun onItemActionClick(holder: RecyclerView.ViewHolder, id: Int, position: Int) {
            val listener = adapterRef?.activityClickListener ?: return
            listener.onStatusActionClick(holder as IStatusViewHolder, id, position)
        }

        override fun onItemActionLongClick(holder: RecyclerView.ViewHolder, id: Int, position: Int): Boolean {
            val listener = adapterRef?.activityClickListener ?: return false
            return listener.onStatusActionLongClick(holder as IStatusViewHolder, id, position)
        }

        override fun onUserProfileClick(holder: IStatusViewHolder, position: Int) {
            val adapter = adapterRef ?: return
            val status = adapter.getActivity(position).activityStatus ?: return
            IntentUtils.openUserProfile(adapter.context, status.account_key, status.user_key,
                    status.user_screen_name, status.user_profile_image_url,
                    PreferencesSingleton.get(adapter.context)[newDocumentApiKey], null)
        }

        override fun onStatusClick(holder: IStatusViewHolder, position: Int) {
            val adapter = adapterRef ?: return
            adapter.activityClickListener?.onStatusClick(holder, position)
        }

        override fun onStatusLongClick(holder: IStatusViewHolder, position: Int): Boolean {
            val listener = adapterRef?.activityClickListener ?: return false
            return listener.onStatusLongClick(holder, position)
        }


        override fun onQuotedStatusClick(holder: IStatusViewHolder, position: Int) {
            val adapter = adapterRef ?: return
            adapter.activityClickListener?.onQuotedStatusClick(holder, position)
        }

        override fun onMediaClick(holder: IStatusViewHolder, view: View, current: ParcelableMedia, statusPosition: Int) {
            val adapter = adapterRef ?: return
            adapter.activityClickListener?.onMediaClick(holder, view, current, statusPosition)
        }

        override fun onActivityClick(holder: ActivityTitleSummaryViewHolder, position: Int) {
            val adapter = adapterRef ?: return
            adapter.activityClickListener?.onActivityClick(holder, position)
        }

        override fun onItemMenuClick(holder: RecyclerView.ViewHolder, menuView: View, position: Int) {
            val adapter = adapterRef ?: return
            adapter.activityClickListener?.onStatusMenuClick(holder as StatusViewHolder, menuView, position)
        }
    }

    companion object {

        const val ITEM_INDEX_ACTIVITY = 0
        const val ITEM_INDEX_LOAD_MORE_INDICATOR = 1

    }
}
