package org.mariotaku.twidere.extension.model

import android.content.Context
import android.net.Uri
import android.text.TextUtils
import org.mariotaku.microblog.library.*
import org.mariotaku.microblog.library.auth.BasicAuthorization
import org.mariotaku.microblog.library.auth.EmptyAuthorization
import org.mariotaku.microblog.library.twitter.TwitterWeb
import org.mariotaku.restfu.RestAPIFactory
import org.mariotaku.restfu.RestRequest
import org.mariotaku.restfu.http.Authorization
import org.mariotaku.restfu.http.Endpoint
import org.mariotaku.restfu.http.MultiValueMap
import org.mariotaku.restfu.oauth.OAuthAuthorization
import org.mariotaku.restfu.oauth.OAuthEndpoint
import org.mariotaku.restfu.oauth.OAuthToken
import org.mariotaku.restfu.oauth2.OAuth2Authorization
import org.mariotaku.twidere.Constants.DEFAULT_TWITTER_API_URL_FORMAT
import org.mariotaku.twidere.annotation.AccountType
import org.mariotaku.twidere.dagger.DependencyHolder
import org.mariotaku.twidere.model.account.cred.*
import org.mariotaku.twidere.singleton.CacheSingleton
import org.mariotaku.twidere.singleton.PreferencesSingleton
import org.mariotaku.twidere.util.HttpClientFactory
import org.mariotaku.twidere.util.InternalTwitterContentUtils
import org.mariotaku.twidere.util.MicroBlogAPIFactory
import org.mariotaku.twidere.util.MicroBlogAPIFactory.sFanfouConstantPool
import org.mariotaku.twidere.util.MicroBlogAPIFactory.sTwitterConstantPool
import org.mariotaku.twidere.util.api.*
import org.mariotaku.twidere.util.media.TwidereMediaDownloader
import org.mariotaku.twidere.util.net.TwidereDns

fun Credentials.getAuthorization(cls: Class<*>?): Authorization {
    if (cls != null) {
        when {
            TwitterWeb::class.java.isAssignableFrom(cls) -> {
                return EmptyAuthorization()
            }
        }
    }
    when (this) {
        is OAuthCredentials -> {
            return OAuthAuthorization(consumer_key, consumer_secret, OAuthToken(access_token,
                    access_token_secret))
        }
        is OAuth2Credentials -> {
            return OAuth2Authorization(access_token)
        }
        is BasicCredentials -> {
            return BasicAuthorization(username, password)
        }
        is EmptyCredentials -> {
            return EmptyAuthorization()
        }
    }
    throw UnsupportedOperationException()
}

fun Credentials.getEndpoint(@AccountType accountType: String?, cls: Class<*>): Endpoint {
    val apiUrlFormat: String
    val noVersionSuffix = this.no_version_suffix
    if (!TextUtils.isEmpty(this.api_url_format)) {
        apiUrlFormat = this.api_url_format
    } else {
        apiUrlFormat = DEFAULT_TWITTER_API_URL_FORMAT
    }
    val domain: String?
    val versionSuffix: String?
    when {
        MicroBlog::class.java.isAssignableFrom(cls) -> {
            domain = "api"
            versionSuffix = if (noVersionSuffix) null else "/1.1/"
        }
        Twitter::class.java.isAssignableFrom(cls) -> {
            domain = "api"
            versionSuffix = if (noVersionSuffix) null else "/1.1/"
        }
        TwitterUpload::class.java.isAssignableFrom(cls) -> {
            domain = "upload"
            versionSuffix = if (noVersionSuffix) null else "/1.1/"
        }
        TwitterOAuth::class.java.isAssignableFrom(cls) -> {
            domain = "api"
            versionSuffix = null
        }
        TwitterOAuth2::class.java.isAssignableFrom(cls) -> {
            domain = "api"
            versionSuffix = null
        }
        TwitterUserStream::class.java.isAssignableFrom(cls) -> {
            domain = "userstream"
            versionSuffix = if (noVersionSuffix) null else "/1.1/"
        }
        TwitterCaps::class.java.isAssignableFrom(cls) -> {
            domain = "caps"
            versionSuffix = null
        }
        FanfouStream::class.java.isAssignableFrom(cls) -> {
            domain = "stream"
            versionSuffix = null
        }
        TwitterWeb::class.java.isAssignableFrom(cls) -> {
            domain = null
            versionSuffix = null
        }
        Mastodon::class.java.isAssignableFrom(cls) -> {
            domain = null
            versionSuffix = null
        }
        MastodonOAuth2::class.java.isAssignableFrom(cls) -> {
            domain = null
            versionSuffix = null
        }
        MastodonStreaming::class.java.isAssignableFrom(cls) -> {
            domain = null
            versionSuffix = null
        }
        else -> throw UnsupportedOperationException("Unsupported class $cls")
    }
    val endpointUrl = MicroBlogAPIFactory.getApiUrl(apiUrlFormat, domain, versionSuffix)
    if (this is OAuthCredentials) {
        val signEndpointUrl = if (same_oauth_signing_url) {
            endpointUrl
        } else when (accountType) {
            AccountType.TWITTER -> MicroBlogAPIFactory.getApiUrl(DEFAULT_TWITTER_API_URL_FORMAT,
                    domain, versionSuffix)
            AccountType.FANFOU -> "http://api.fanfou.com/"
            else -> endpointUrl
        }
        return OAuthEndpoint(endpointUrl, signEndpointUrl)
    }
    return Endpoint(endpointUrl)
}

fun <T> Credentials.newMicroBlogInstance(context: Context, @AccountType accountType: String? = null,
        cls: Class<T>): T {
    return newMicroBlogInstance(context, getEndpoint(accountType, cls), getAuthorization(cls), accountType, cls)
}

fun <T> newMicroBlogInstance(context: Context, endpoint: Endpoint, auth: Authorization,
        @AccountType accountType: String? = null, cls: Class<T>): T {
    val factory = RestAPIFactory<MicroBlogException>()
    val extraHeaders = run {
        if (auth !is OAuthAuthorization) return@run null
        val officialKeyType = InternalTwitterContentUtils.getOfficialKeyType(context,
                auth.consumerKey, auth.consumerSecret)
        return@run MicroBlogAPIFactory.getExtraHeaders(context, officialKeyType)
    } ?: UserAgentExtraHeaders(MicroBlogAPIFactory.getTwidereUserAgent(context))
    val holder = DependencyHolder.get(context)
    var extraRequestParams: Map<String, String>? = null
    when (cls) {
        TwitterUpload::class.java -> {
            val conf = HttpClientFactory.HttpClientConfiguration(PreferencesSingleton.get(context))
            // Use longer timeout for uploading
            conf.readTimeoutSecs = 30
            conf.writeTimeoutSecs = 30
            conf.connectionTimeoutSecs = 60
            val dns = TwidereDns.get(context)
            val cache = CacheSingleton.get(context)
            val uploadHttpClient = HttpClientFactory.createRestHttpClient(conf, dns, cache)
            factory.setHttpClient(uploadHttpClient)
        }
        TwitterUserStream::class.java, FanfouStream::class.java, MastodonStreaming::class.java -> {
            val conf = HttpClientFactory.HttpClientConfiguration(PreferencesSingleton.get(context))
            // Use longer read timeout for streaming
            conf.readTimeoutSecs = 300
            val dns = TwidereDns.get(context)
            val cache = CacheSingleton.get(context)
            val streamHttpClient = HttpClientFactory.createRestHttpClient(conf, dns, cache)
            factory.setHttpClient(streamHttpClient)
        }
        else -> {
            factory.setHttpClient(holder.restHttpClient)
        }
    }
    factory.setAuthorization(auth)
    factory.setEndpoint(endpoint)
    when (accountType) {
        AccountType.TWITTER -> {
            factory.setConstantPool(sTwitterConstantPool)
        }
        AccountType.FANFOU -> {
            factory.setConstantPool(sFanfouConstantPool)
            if (cls != FanfouStream::class.java) {
                extraRequestParams = mapOf("format" to "html")
            }
        }
    }
    factory.setRestConverterFactory(TwitterConverterFactory)
    factory.setExceptionFactory(TwidereExceptionFactory)
    factory.setRestRequestFactory(TwidereRestRequestFactory(extraRequestParams))
    factory.setHttpRequestFactory(TwidereHttpRequestFactory(extraHeaders))
    return factory.build<T>(cls)
}

internal fun Credentials.authorizationHeader(
        uri: Uri,
        modifiedUri: Uri = TwidereMediaDownloader.getReplacedUri(uri, api_url_format) ?: uri,
        cls: Class<*>? = null
): String {
    val auth = getAuthorization(cls)
    val endpoint: Endpoint
    if (auth is OAuthAuthorization) {
        endpoint = OAuthEndpoint(TwidereMediaDownloader.getEndpoint(modifiedUri),
                TwidereMediaDownloader.getEndpoint(uri))
    } else {
        endpoint = Endpoint(TwidereMediaDownloader.getEndpoint(modifiedUri))
    }
    val queries = MultiValueMap<String>()
    for (name in uri.queryParameterNames) {
        for (value in uri.getQueryParameters(name)) {
            queries.add(name, value)
        }
    }
    val info = RestRequest("GET", false, uri.path, null, queries, null, null, null, null)
    return auth.getHeader(endpoint, info)
}