/*
 *             Twidere - Twitter client for Android
 *
 *  Copyright (C) 2012-2017 Mariotaku Lee <mariotaku.lee@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.mariotaku.twidere.extension.model.api

import android.text.Spanned
import android.text.style.URLSpan
import org.apache.commons.text.translate.LookupTranslator
import org.mariotaku.commons.text.CodePointArray
import org.mariotaku.ktextension.isNotNullOrEmpty
import org.mariotaku.ktextension.mapToArray
import org.mariotaku.microblog.library.model.microblog.*
import org.mariotaku.twidere.exception.MalformedResponseException
import org.mariotaku.twidere.extension.model.addFilterFlag
import org.mariotaku.twidere.extension.model.quoted
import org.mariotaku.twidere.extension.model.updateContentFilterInfo
import org.mariotaku.twidere.extension.model.updateFilterInfo
import org.mariotaku.twidere.extension.toSpanItem
import org.mariotaku.twidere.model.*
import org.mariotaku.twidere.model.attachment.QuotedStatus
import org.mariotaku.twidere.model.attachment.SummaryCard
import org.mariotaku.twidere.model.util.ParcelableMediaUtils
import org.mariotaku.twidere.text.placeholder.CustomEmojiShortCodeSpan
import org.mariotaku.twidere.util.EntityArrays
import org.mariotaku.twidere.util.HtmlBuilder
import org.mariotaku.twidere.util.HtmlSpanBuilder

fun Status.toParcelable(details: AccountDetails, statusConfig: ModelCreationConfig = ModelCreationConfig.DEFAULT,
        updateFilterInfoAction: (Status, ParcelableStatus) -> Unit = ::updateFilterInfoDefault): ParcelableStatus {
    return toParcelable(details.key, details.type, statusConfig, updateFilterInfoAction).apply {
        account_color = details.color
    }
}

fun Status.toParcelable(accountKey: UserKey, accountType: String, statusConfig: ModelCreationConfig = ModelCreationConfig.DEFAULT,
        updateFilterInfoAction: (Status, ParcelableStatus) -> Unit = ::updateFilterInfoDefault): ParcelableStatus {
    val result = ParcelableStatus()
    applyTo(accountKey, accountType, statusConfig, result, updateFilterInfoAction)
    return result
}

fun Status.applyTo(accountKey: UserKey, accountType: String, statusConfig: ModelCreationConfig = ModelCreationConfig.DEFAULT,
        result: ParcelableStatus,
        updateFilterInfoAction: (Status, ParcelableStatus) -> Unit = ::updateFilterInfoDefault) {
    val extras = ParcelableStatus.Extras()
    val attachment = ParcelableStatusAttachment()
    result.account_key = accountKey
    result.id = id
    result.sort_id = sortId
    result.timestamp = createdAt?.time ?: 0

    extras.external_url = inferredExternalUrl
    extras.support_entities = entities != null
    extras.statusnet_conversation_id = statusnetConversationId
    extras.conversation_id = conversationId
    result.is_pinned_status = user.pinnedTweetIds?.contains(id) ?: false

    val retweetedStatus = retweetedStatus
    result.is_retweet = isRetweet
    result.retweeted = wasRetweeted()
    val status: Status
    if (retweetedStatus != null) {
        status = retweetedStatus
        val retweetUser = this.user
        result.retweet_id = retweetedStatus.id
        result.retweet_timestamp = retweetedStatus.createdAt?.time ?: 0
        result.retweeted_by_user_key = retweetUser.key
        result.retweeted_by_user_name = retweetUser.name
        result.retweeted_by_user_screen_name = retweetUser.screenName
        result.retweeted_by_user_profile_image = retweetUser.getProfileImageOfSize(statusConfig.profileImageSize)

        extras.retweeted_external_url = retweetedStatus.inferredExternalUrl
    } else {
        status = this
    }

    val user = status.user ?: throw MalformedResponseException()
    if (user.isBlocking == true) {
        result.addFilterFlag(ParcelableStatus.FilterFlags.BLOCKING_USER)
    }
    if (user.isBlockedBy == true) {
        result.addFilterFlag(ParcelableStatus.FilterFlags.BLOCKED_BY_USER)
    }
    if (status.isPossiblySensitive) {
        result.addFilterFlag(ParcelableStatus.FilterFlags.POSSIBLY_SENSITIVE)
    }

    val quotedStatus = status.quotedStatus
    result.is_quote = status.isQuoteStatus
    if (quotedStatus != null) {
        val quoted = QuotedStatus()
        val quotedUser = quotedStatus.user ?: throw MalformedResponseException()
        quoted.id = quotedStatus.id
        quoted.account_key = accountKey
        quoted.external_url = quotedStatus.inferredExternalUrl

        val quotedText = quotedStatus.htmlText
        // Twitter will escape <> to &lt;&gt;, so if a status contains those symbols unescaped
        // We should treat this as an html
        if (quotedText.isHtml) {
            val html = HtmlSpanBuilder.fromHtml(quotedText, quotedStatus.extendedText)
            quoted.text_unescaped = html?.toString()
            quoted.text_plain = quoted.text_unescaped
            quoted.spans = html?.spanItems
        } else {
            val textWithIndices = quotedStatus.formattedTextWithIndices()
            quoted.text_plain = quotedText.twitterUnescaped()
            quoted.text_unescaped = textWithIndices.text
            quoted.spans = textWithIndices.spans
            extras.quoted_display_text_range = textWithIndices.range
        }

        quoted.timestamp = quotedStatus.createdAt.time
        quoted.source = quotedStatus.source
        quoted.media = ParcelableMediaUtils.fromStatus(quotedStatus, accountKey, accountType)

        quoted.user_key = quotedUser.key
        quoted.user_name = quotedUser.name
        quoted.user_screen_name = quotedUser.screenName
        quoted.user_profile_image = quotedUser.getProfileImageOfSize(statusConfig.profileImageSize)
        quoted.user_is_protected = quotedUser.isProtected
        quoted.user_is_verified = quotedUser.isVerified

        if (quotedStatus.isPossiblySensitive) {
            result.addFilterFlag(ParcelableStatus.FilterFlags.POSSIBLY_SENSITIVE)
        }

        attachment.quoted = quoted
    } else if (status.isQuoteStatus) {
        result.addFilterFlag(ParcelableStatus.FilterFlags.QUOTE_NOT_AVAILABLE)
    }

    result.reply_count = status.replyCount
    result.retweet_count = status.retweetCount
    result.favorite_count = status.favoriteCount

    result.in_reply_to_name = status.inReplyToName
    result.in_reply_to_screen_name = status.inReplyToScreenName
    result.in_reply_to_status_id = status.inReplyToStatusId
    result.in_reply_to_user_key = status.getInReplyToUserKey(accountKey)

    result.user_key = user.key
    result.user_name = user.name
    result.user_screen_name = user.screenName
    result.user_profile_image_url = user.getProfileImageOfSize(statusConfig.profileImageSize)
    result.user_is_protected = user.isProtected
    result.user_is_verified = user.isVerified
    result.user_is_following = user.isFollowing == true
    val text = status.htmlText
    // Twitter will escape <> to &lt;&gt;, so if a status contains those symbols unescaped
    // We should treat this as an html
    if (text.isHtml) {
        val html = HtmlSpanBuilder.fromHtml(text, status.extendedText)
        result.text_unescaped = html?.toString()
        result.text_plain = result.text_unescaped
        result.spans = html?.spanItems
    } else {
        val textWithIndices = status.formattedTextWithIndices()
        result.text_unescaped = textWithIndices.text
        result.text_plain = text.twitterUnescaped()
        result.spans = textWithIndices.spans
        extras.display_text_range = textWithIndices.range
    }

    attachment.media = ParcelableMediaUtils.fromStatus(status, accountKey, accountType)
    when (status.card?.name) {
        "summary" -> {
            attachment.summary_card = status.card?.toSummaryCard(statusConfig.summaryThumbnailSize)
        }
    }

    result.source = status.source
    result.location = status.parcelableLocation
    result.is_favorite = status.isFavorited
    if (result.account_key.maybeEquals(result.retweeted_by_user_key)) {
        result.my_retweet_id = result.id
    } else {
        result.my_retweet_id = status.currentUserRetweet
    }
    result.is_possibly_sensitive = status.isPossiblySensitive
    result.mentions = status.userMentionEntities?.mapToArray { it.toParcelable(user.host) }

    result.card_name = result.attachment?.card?.name
    result.place_full_name = status.placeFullName
    result.lang = status.lang
    result.extras = extras

    result.attachment = attachment

    if (result.attachment?.media.isNotNullOrEmpty() || result.quoted?.media.isNotNullOrEmpty()) {
        result.addFilterFlag(ParcelableStatus.FilterFlags.HAS_MEDIA)
    }
    if (user.isFollowing == false) {
        result.addFilterFlag(ParcelableStatus.FilterFlags.NOT_FRIEND)
    }
    if (user.isFollowedBy == false) {
        result.addFilterFlag(ParcelableStatus.FilterFlags.NOT_FOLLOWER)
    }

    updateFilterInfoAction(this, result)
}


fun Status.formattedTextWithIndices(): StatusTextWithIndices {
    val source = CodePointArray(this.fullText ?: this.text!!)
    val builder = HtmlBuilder(source, false, true, false)
    builder.addEntities(this)
    val textWithIndices = StatusTextWithIndices()
    val (text, spans) = builder.buildWithIndices()
    textWithIndices.text = text
    textWithIndices.spans = spans

    // Display text range
    val range = displayTextRange?.takeIf { it.size == 2 }
    if (range != null) {
        textWithIndices.range = intArrayOf(
                source.findResultRangeLength(spans, 0, range[0]),
                text.length - source.findResultRangeLength(spans, range[1], source.length())
        )
    }
    return textWithIndices
}

fun CodePointArray.findResultRangeLength(spans: Array<SpanItem>, origStart: Int, origEnd: Int): Int {
    val findResult = findByOrigRange(spans, origStart, origEnd)
    if (findResult.isEmpty()) {
        return charCount(origStart, origEnd)
    }
    val first = findResult.first()
    val last = findResult.last()
    if (first.orig_start == -1 || last.orig_end == -1)
        return charCount(origStart, origEnd)
    return charCount(origStart, first.orig_start) + (last.end - first.start) + charCount(first.orig_end, origEnd)
}

fun HtmlBuilder.addEntities(entities: EntitySupport) {
    // Format media.
    var mediaEntities: Array<MediaEntity>? = null
    if (entities is ExtendedEntitySupport) {
        mediaEntities = entities.extendedMediaEntities
    }
    if (mediaEntities == null) {
        mediaEntities = entities.mediaEntities
    }
    val startEnd = IntArray(2)
    mediaEntities?.forEach { mediaEntity ->
        val mediaUrl = mediaEntity.bestMediaUrl
        val expandedUrl = mediaEntity.expandedUrl
        val displayUrl = mediaEntity.displayUrl
        if (mediaUrl != null && displayUrl != null && mediaEntity.getStartEndForEntity(startEnd)) {
            addLink(expandedUrl, displayUrl,
                    startEnd[0], startEnd[1], false)
        }
    }
    entities.urlEntities?.forEach { urlEntity ->
        val expandedUrl = urlEntity.expandedUrl
        val displayUrl = urlEntity.displayUrl
        if (expandedUrl != null && displayUrl != null && urlEntity.getStartEndForEntity(startEnd)) {
            addLink(expandedUrl, displayUrl, startEnd[0],
                    startEnd[1], false)
        }
    }
}


fun updateFilterInfoDefault(status: Status, result: ParcelableStatus) {
    result.updateFilterInfo(setOf(
            status.userDescriptionUnescaped,
            status.userUrlExpanded,
            status.userLocation,
            status.retweetedStatus?.userDescriptionUnescaped,
            status.retweetedStatus?.userLocation,
            status.retweetedStatus?.userUrlExpanded,
            status.quotedStatus?.userDescriptionUnescaped,
            status.quotedStatus?.userLocation,
            status.quotedStatus?.userUrlExpanded
    ))
}

/**
 * Ignores status user info
 */
fun updateFilterInfoForUserTimeline(status: Status, result: ParcelableStatus) {
    result.updateContentFilterInfo()

    if (result.is_retweet) {
        result.filter_users = setOf(result.user_key, result.quoted?.user_key).filterNotNull().toTypedArray()
        result.filter_names = setOf(result.user_name, result.quoted?.user_name).filterNotNull().toTypedArray()
        result.filter_descriptions = setOf(
                status.retweetedStatus?.userDescriptionUnescaped,
                status.retweetedStatus?.userUrlExpanded,
                status.retweetedStatus?.userLocation,
                status.quotedStatus?.userDescriptionUnescaped,
                status.quotedStatus?.userLocation,
                status.quotedStatus?.userUrlExpanded
        ).filterNotNull().joinToString("\n")
    } else {
        result.filter_users = setOf(result.quoted?.user_key).filterNotNull().toTypedArray()
        result.filter_names = setOf(result.quoted?.user_name).filterNotNull().toTypedArray()
        result.filter_descriptions = setOf(
                status.quotedStatus?.userDescriptionUnescaped,
                status.quotedStatus?.userLocation,
                status.quotedStatus?.userUrlExpanded
        ).filterNotNull().joinToString("\n")
    }
}

private fun CardEntity.toSummaryCard(thumbnailSize: Int): SummaryCard? {
    val values = bindingValues ?: return null
    val obj = SummaryCard()
    obj.title = (values["title"] as? CardEntity.StringValue)?.value ?: return null
    obj.description = (values["description"] as? CardEntity.StringValue)?.value
    obj.domain = (values["domain"] as? CardEntity.StringValue)?.value
    obj.url = url
    val thumbnails = values.mapNotNull { (k, v) ->
        if (!k.startsWith("thumbnail_")) return@mapNotNull null
        return@mapNotNull v as? CardEntity.ImageValue
    }.sortedBy { Math.min(it.width, it.height) }
    val bestThumbnail = thumbnails.firstOrNull {
        Math.min(it.width, it.height) >= thumbnailSize
    } ?: thumbnails.lastOrNull()
    obj.thumbnail = bestThumbnail?.url
    return obj
}

private fun String.twitterUnescaped(): String {
    return TwitterRawTextTranslator.translate(this)
}

private inline val Status.userDescriptionUnescaped: String?
    get() = user?.let { it.formatUserDescription()?.first }

private inline val Status.userUrlExpanded: String?
    get() = user?.urlEntities?.firstOrNull()?.expandedUrl

private inline val Status.userLocation: String?
    get() = user?.location

/**
 * @param spans Ordered spans
 * *
 * @param start orig_start
 * *
 * @param end   orig_end
 */
internal fun findByOrigRange(spans: Array<SpanItem>, start: Int, end: Int): List<SpanItem> {
    return spans.filter { it.orig_start >= start && it.orig_end <= end }
}

internal inline val CharSequence.spanItems
    get() = (this as? Spanned)?.let { text ->
        text.getSpans(0, length, Any::class.java).mapNotNull { span ->
            return@mapNotNull when (span) {
                is URLSpan -> span.toSpanItem(text)
                is CustomEmojiShortCodeSpan -> span.toSpanItem(text)
                else -> null
            }
        }.toTypedArray()
    }

internal inline val String.isHtml get() = contains('<') && contains('>')

private inline val Status.inReplyToName
    get() = userMentionEntities?.firstOrNull {
        inReplyToUserId == it.id
    }?.name ?: attentions?.firstOrNull {
        inReplyToUserId == it.id
    }?.fullName ?: inReplyToScreenName


private inline val Status.placeFullName
    get() = place?.fullName ?: location?.takeIf {
        ParcelableLocation.valueOf(location) == null
    }

private inline val Status.inferredExternalUrl
    get() = externalUrl ?: uri?.let { uri ->
        noticeUriRegex.matchEntire(uri)?.let { result: MatchResult ->
            "https://${result.groups[1]?.value}/notice/${result.groups[3]?.value}"
        }
    }

private val Status.parcelableLocation: ParcelableLocation?
    get() {
        val geoLocation = geoLocation
        if (geoLocation != null) {
            return ParcelableLocation(geoLocation.latitude, geoLocation.longitude)
        }
        val locationString = location ?: return null
        val location = ParcelableLocation.valueOf(locationString)
        if (location != null) {
            return location
        }
        return null
    }

private fun Status.getInReplyToUserKey(accountKey: UserKey): UserKey? {
    val inReplyToUserId = inReplyToUserId ?: return null
    val entities = userMentionEntities
    if (entities != null) {
        if (entities.any { inReplyToUserId == it.id }) {
            return UserKey(inReplyToUserId, accountKey.host)
        }
    }
    val attentions = attentions
    if (attentions != null) {
        attentions.firstOrNull { inReplyToUserId == it.id }?.let {
            val host = getUserHost(it.ostatusUri, accountKey.host)
            return UserKey(inReplyToUserId, host)
        }
    }
    return UserKey(inReplyToUserId, accountKey.host)
}

private val noticeUriRegex = Regex("tag:([\\w\\d.]+),(\\d{4}-\\d{2}-\\d{2}):noticeId=(\\d+):objectType=(\\w+)")

private object TwitterRawTextTranslator : LookupTranslator(EntityArrays.BASIC_UNESCAPE)

class StatusTextWithIndices {
    var text: String? = null
    var spans: Array<SpanItem>? = null
    var range: IntArray? = null
}