/*
 * 				Twidere - Twitter client for Android
 * 
 *  Copyright (C) 2012-2014 Mariotaku Lee <mariotaku.lee@gmail.com>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.mariotaku.twidere.fragment

import android.app.Dialog
import android.content.ContentValues
import android.os.Bundle
import android.support.v4.app.FragmentManager
import android.support.v7.app.AlertDialog
import com.twitter.Extractor
import org.mariotaku.ktextension.mapToArray
import org.mariotaku.twidere.R
import org.mariotaku.twidere.constant.IntentConstants.EXTRA_STATUS
import org.mariotaku.twidere.extension.applyTheme
import org.mariotaku.twidere.extension.blockBulkInsert
import org.mariotaku.twidere.extension.bulkDelete
import org.mariotaku.twidere.extension.model.quoted
import org.mariotaku.twidere.extension.onShow
import org.mariotaku.twidere.model.ParcelableStatus
import org.mariotaku.twidere.model.ParcelableUserMention
import org.mariotaku.twidere.model.UserKey
import org.mariotaku.twidere.provider.TwidereDataStore.Filters
import org.mariotaku.twidere.util.ContentValuesCreator
import org.mariotaku.twidere.util.HtmlEscapeHelper
import org.mariotaku.twidere.util.UserColorNameManager
import java.util.*

class AddStatusFilterDialogFragment : BaseDialogFragment() {

    private val extractor = Extractor()
    private val filterItemsInfo: Array<FilterItemInfo>
        get() {
            val status = arguments!!.getParcelable<ParcelableStatus>(EXTRA_STATUS)
                    ?: return emptyArray()
            val list = ArrayList<FilterItemInfo>()
            if (status.is_retweet && status.retweeted_by_user_key != null) {
                list.add(FilterItemInfo(FilterItemInfo.FILTER_TYPE_USER,
                        UserItem(status.retweeted_by_user_key!!, status.retweeted_by_user_name,
                                status.retweeted_by_user_screen_name)))
            }
            if (status.is_quote && status.quoted?.user_key != null) {
                list.add(FilterItemInfo(FilterItemInfo.FILTER_TYPE_USER,
                        UserItem(status.quoted?.user_key!!, status.quoted?.user_name!!,
                                status.quoted?.user_screen_name!!)))
            }
            list.add(FilterItemInfo(FilterItemInfo.FILTER_TYPE_USER, UserItem(status.user_key,
                    status.user_name, status.user_screen_name)))
            val mentions = status.mentions
            mentions?.filter { it.key != status.user_key }?.mapTo(list) {
                FilterItemInfo(FilterItemInfo.FILTER_TYPE_USER, it)
            }
            val hashtags = HashSet<String>()
            hashtags.addAll(extractor.extractHashtags(status.text_plain))
            for (hashtag in hashtags) {
                list.add(FilterItemInfo(FilterItemInfo.FILTER_TYPE_KEYWORD, hashtag))
            }
            val source = status.source?.let(HtmlEscapeHelper::toPlainText)
            if (source != null) {
                list.add(FilterItemInfo(FilterItemInfo.FILTER_TYPE_SOURCE, source))
            }
            return list.toTypedArray()
        }

    override fun onCreateDialog(savedInstanceState: Bundle?): Dialog {
        val manager = UserColorNameManager.get(context!!)
        val builder = AlertDialog.Builder(context!!)
        val filterItems = filterItemsInfo
        val entries = filterItems.mapToArray { info ->
            when (info.type) {
                FilterItemInfo.FILTER_TYPE_USER -> getString(R.string.user_filter_name,
                        getName(manager, info.value))
                FilterItemInfo.FILTER_TYPE_KEYWORD -> getString(R.string.keyword_filter_name,
                        getName(manager, info.value))
                FilterItemInfo.FILTER_TYPE_SOURCE -> getString(R.string.source_filter_name,
                        getName(manager, info.value))
                else -> throw UnsupportedOperationException()
            }
        }
        builder.setTitle(R.string.action_add_to_filter)
        builder.setMultiChoiceItems(entries, null, null)
        builder.setPositiveButton(android.R.string.ok) { dialog, _ ->
            val alertDialog = dialog as AlertDialog
            val checkPositions = alertDialog.listView.checkedItemPositions

            val userKeys = HashSet<UserKey>()
            val keywords = HashSet<String>()
            val sources = HashSet<String>()
            val userValues = ArrayList<ContentValues>()
            val keywordValues = ArrayList<ContentValues>()
            val sourceValues = ArrayList<ContentValues>()
            loop@ for (i in 0 until checkPositions.size()) {
                if (!checkPositions.valueAt(i)) {
                    continue@loop
                }
                val info = filterItems[checkPositions.keyAt(i)]
                val value = info.value
                if (value is ParcelableUserMention) {
                    userKeys.add(value.key)
                    userValues.add(ContentValuesCreator.createFilteredUser(value))
                } else if (value is UserItem) {
                    userKeys.add(value.key)
                    userValues.add(createFilteredUser(value))
                } else if (info.type == FilterItemInfo.FILTER_TYPE_KEYWORD) {
                    val keyword = value.toString()
                    keywords.add(keyword)
                    val values = ContentValues()
                    values.put(Filters.Keywords.VALUE, "#$keyword")
                    keywordValues.add(values)
                } else if (info.type == FilterItemInfo.FILTER_TYPE_SOURCE) {
                    val source = value.toString()
                    sources.add(source)
                    val values = ContentValues()
                    values.put(Filters.Sources.VALUE, source)
                    sourceValues.add(values)
                }
            }
            val resolver = context!!.contentResolver
            resolver.bulkDelete(Filters.Users.CONTENT_URI, Filters.Users.USER_KEY, userKeys.mapToArray(UserKey::toString))
            resolver.bulkDelete(Filters.Keywords.CONTENT_URI, Filters.Keywords.VALUE, keywords.toTypedArray())
            resolver.bulkDelete(Filters.Sources.CONTENT_URI, Filters.Sources.VALUE, sources.toTypedArray())
            resolver.blockBulkInsert(Filters.Users.CONTENT_URI, userValues)
            resolver.blockBulkInsert(Filters.Keywords.CONTENT_URI, keywordValues)
            resolver.blockBulkInsert(Filters.Sources.CONTENT_URI, sourceValues)
        }
        builder.setNegativeButton(android.R.string.cancel, null)
        val dialog = builder.create()
        dialog.onShow { it.applyTheme() }
        return dialog
    }

    private fun getName(manager: UserColorNameManager, value: Any): String {
        return when (value) {
            is ParcelableUserMention -> manager.getDisplayName(value.key, value.name, value.screen_name)
            is UserItem -> manager.getDisplayName(value.key, value.name, value.screen_name)
            else -> value.toString()
        }
    }

    internal data class FilterItemInfo(
            val type: Int,
            val value: Any
    ) {

        companion object {

            internal const val FILTER_TYPE_USER = 1
            internal const val FILTER_TYPE_KEYWORD = 2
            internal const val FILTER_TYPE_SOURCE = 3
        }

    }

    internal data class UserItem(
            val key: UserKey,
            val name: String,
            val screen_name: String
    )

    companion object {

        const val FRAGMENT_TAG = "add_status_filter"

        private fun createFilteredUser(item: UserItem): ContentValues {
            val values = ContentValues()
            values.put(Filters.Users.USER_KEY, item.key.toString())
            values.put(Filters.Users.NAME, item.name)
            values.put(Filters.Users.SCREEN_NAME, item.screen_name)
            return values
        }

        fun show(fm: FragmentManager, status: ParcelableStatus): AddStatusFilterDialogFragment {
            val args = Bundle()
            args.putParcelable(EXTRA_STATUS, status)
            val f = AddStatusFilterDialogFragment()
            f.arguments = args
            f.show(fm, FRAGMENT_TAG)
            return f
        }
    }

}
