/*
 *             Twidere - Twitter client for Android
 *
 *  Copyright (C) 2012-2017 Mariotaku Lee <mariotaku.lee@gmail.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.mariotaku.twidere.promise

import android.app.Application
import nl.komponents.kovenant.Promise
import nl.komponents.kovenant.combine.and
import nl.komponents.kovenant.then
import nl.komponents.kovenant.ui.alwaysUi
import nl.komponents.kovenant.ui.promiseOnUi
import nl.komponents.kovenant.ui.successUi
import org.mariotaku.microblog.library.Mastodon
import org.mariotaku.microblog.library.MicroBlog
import org.mariotaku.microblog.library.MicroBlogException
import org.mariotaku.microblog.library.model.microblog.ErrorInfo
import org.mariotaku.twidere.R
import org.mariotaku.twidere.annotation.AccountType
import org.mariotaku.twidere.extension.model.api.mastodon.toParcelable
import org.mariotaku.twidere.extension.model.api.toParcelable
import org.mariotaku.twidere.extension.model.newMicroBlogInstance
import org.mariotaku.twidere.extension.promise
import org.mariotaku.twidere.extension.promise.accountTask
import org.mariotaku.twidere.extension.promise.toData
import org.mariotaku.twidere.extension.promise.toastOnResult
import org.mariotaku.twidere.model.ParcelableStatus
import org.mariotaku.twidere.model.UserKey
import org.mariotaku.twidere.model.event.StatusDestroyedEvent
import org.mariotaku.twidere.model.event.StatusListChangedEvent
import org.mariotaku.twidere.singleton.BusSingleton
import org.mariotaku.twidere.task.AbsAccountRequestTask
import org.mariotaku.twidere.task.CreateFavoriteTask
import org.mariotaku.twidere.task.DestroyFavoriteTask
import org.mariotaku.twidere.task.RetweetStatusTask
import org.mariotaku.twidere.util.deleteActivityStatus
import org.mariotaku.twidere.util.deleteStatus
import org.mariotaku.twidere.util.lang.ApplicationContextSingletonHolder

class StatusPromises private constructor(private val application: Application) {

    fun destroy(accountKey: UserKey, id: String): Promise<ParcelableStatus, Exception> = (promiseOnUi {
        DestroyTasks.addTaskId(accountKey, id)
        BusSingleton.post(StatusListChangedEvent())
    } and accountTask(application, accountKey) { account ->
        when (account.type) {
            AccountType.MASTODON -> {
                val mastodon = account.newMicroBlogInstance(application, cls = Mastodon::class.java)
                val result = mastodon.fetchStatus(id)
                mastodon.deleteStatus(id)
                return@accountTask result.toParcelable(account)
            }
            else -> {
                val microBlog = account.newMicroBlogInstance(application, cls = MicroBlog::class.java)
                return@accountTask microBlog.destroyStatus(id).toParcelable(account)
            }
        }
    }).then { (_, status) ->
        return@then status
    }.success { status ->
        application.contentResolver.deleteStatus(accountKey, id, status)
        application.contentResolver.deleteActivityStatus(accountKey, id, status)
    }.fail { ex ->
        if (ex is MicroBlogException && ex.errorCode == ErrorInfo.STATUS_NOT_FOUND) {
            application.contentResolver.deleteStatus(accountKey, id, null)
            application.contentResolver.deleteActivityStatus(accountKey, id, null)
        }
    }.alwaysUi {
        DestroyTasks.removeTaskId(accountKey, id)
    }.successUi { status ->
        BusSingleton.post(StatusDestroyedEvent(status))
    }.toastOnResult(application) { status ->
        if (status.retweet_id != null) {
            return@toastOnResult application.getString(R.string.message_toast_retweet_cancelled)
        } else {
            return@toastOnResult application.getString(R.string.message_toast_status_deleted)
        }
    }

    fun cancelRetweet(accountKey: UserKey, statusId: String?, myRetweetId: String?): Promise<ParcelableStatus, Exception> = when {
        myRetweetId != null -> destroy(accountKey, myRetweetId)
        statusId != null -> destroy(accountKey, statusId)
        else -> Promise.ofFail(IllegalArgumentException())
    }

    fun retweet(accountKey: UserKey, status: ParcelableStatus): Promise<ParcelableStatus, Exception> =
            RetweetStatusTask(application, accountKey, status).promise().toData()

    fun favorite(accountKey: UserKey, status: ParcelableStatus): Promise<ParcelableStatus, Exception> =
            CreateFavoriteTask(application, accountKey, status).promise().toData()

    fun unfavorite(accountKey: UserKey, id: String): Promise<ParcelableStatus, Exception> =
            DestroyFavoriteTask(application, accountKey, id).promise().toData()

    companion object : ApplicationContextSingletonHolder<StatusPromises>(::StatusPromises)

    private object DestroyTasks : AbsAccountRequestTask.ObjectIdTaskCompanion()
}